#include <itf/option.h>
#include <itf/itf.h>
#include <sim/sim.h>


int command_build(tokc,tokv)
int tokc;
char *tokv[];
{
  int n, i, n_layers;
  int input_shift = 0, output_shift = 0;
  int encode_input = 0, share_encoding = 0;
  int squashing_function = -1;
  int autobias = 1;
  int input_bias = 0;
  int restrict = 0;
  int seed = NULL;
  char *name,*builder;
  static struct OPTION options[] =
    {{OPTPRES,          "sigmoid",	0},
     {OPTPRES,          "tanh",		0},
     {OPTSIGN,          "autobias",	0},
     {OPTSIGN,          "input-bias",	0},
     {OPTPRES|OPTINTARG,"restrict",	0},
     {OPTPRES|OPTINTARG,"seed",		0},
     {OPTPRES|OPTSTRARG,"name",		0},
     {OPTPRES|OPTINTARG,"encode-input",	0},
     {OPTPRES,          "share-encoding",0},
     {OPTPRES|OPTINTARG,"input-shift",	0},
     {OPTPRES|OPTINTARG,"output-shift",	0},
     {OPTLAST,NULL,0}};

  IUsage("[%pseed <int>] [%psigmoid] [%ptanh] [%sautobias] [%sinput-bias]\
 [%prestrict <int>] [%pencode-input <#blocks>] [%pshare-encoding]\
 [%pinput-shift <N>] [%poutput-shift <N>] [<name>] <size1> <size2> ...");
  if (GiveHelp(tokc)) {
    ISynopsis("build a new backprop network");
    IHelp(IHelpArgs,
"Build a backpropagation network with the given number of layers, and",
"with the given number of units in each layer.",
"The first layer will be the input layer, the last the output layer.",
"The zeroeth layer will be built for the bias unit.",
"For example:",
"  build foo 2 2 1",
"will build a network named \"foo\" with four layers (one extra for the",
"bias unit).  There will be two units in the first layer, two in the",
"second, and one in the third.  Layer 1 will be the input layer, layer",
"3 will be the output layer, and layer 0 will be the bias layer.",
"OPTIONS",
"%pseed <int> : Seed random number generator with <int>.",
"%psigmoid : use the sigmoid squashing function (the default).",
"%ptanh : use the tanh squashing function.",
"%sautobias : automatically connect the bias (default %y).",
"%sinput-bias : connect the input layer to the bias (default %n).",
"%prestrict <int> : Restrict biases to be non-negative (arg of 1)",
"  or non-positive (arg of -1) (default 0).",
"%poutput-shift <N> : divide the output layer into <N> shift buffers.",
"  When a new target is loaded, the previous ones will be shifted.",
"%pinput-shift <N> : divide the input layer into <N> shift buffers.",
"  When a new input is loaded, the previous ones will be shifted.",
"%pencode-input <N>: divide the input into N blocks and connect the",
"  first N hidden layers to each of these blocks.  If \"input-shift\"",
"  is also used, N should be the same for both.",
"%pshare-encoding : share the links created by \"encode-input\".",
	  NULL);
    return;
  }
  if (tokc<2) {
    IError(IPrintUsage(tokv[0],usage));
    return;
  }
  /***
   * process options, putting some of them (not the name) in the builder string
   */
  builder = IComposeTokvLine(tokv,1);
  tokv++;
  tokc--;
  name = NULL;
  while (IsOption(tokv[0])) {
    if (OptionMatch(tokv[0],"tanh",options,NULL)) {
      squashing_function = 1;
     } else if (OptionMatch(tokv[0],"seed",options,NULL)) {
       if (!IExpectInteger(tokv[1])) return;
       seed = atoi(*++tokv); tokc--;
     } else if (OptionMatch(tokv[0],"restrict",options,NULL)) {
       if (!IExpectInteger(tokv[1])) return;
       restrict = atoi(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"sigmoid",options,NULL)) {
      squashing_function = 0;
    } else if (OptionMatch(tokv[0],"autobias",options,&autobias)) {
    } else if (OptionMatch(tokv[0],"input-bias",options,&input_bias)) {
    } else if (OptionMatch(tokv[0],"name",options,NULL)) {
      if (!IExpectVarName(tokv[1])) return;
      name = *++tokv; tokc--;
    } else if (OptionMatch(tokv[0],"input-shift",options,NULL)) {
      if (!IExpectInteger(tokv[1])) return;
       input_shift = atoi(*++tokv); tokc--;
     } else if (OptionMatch(tokv[0],"output-shift",options,NULL)) {
       if (!IExpectInteger(tokv[1])) return;
       output_shift = atoi(*++tokv); tokc--;
     } else if (OptionMatch(tokv[0],"encode-input",options,NULL)) {
       if (!IExpectInteger(tokv[1])) return;
       encode_input = atoi(*++tokv); tokc--;
     } else if (OptionMatch(tokv[0],"share-encoding",options,&share_encoding)) {
     } else {
       IError("option not understood: \"%s\"",tokv[0]);
       return;
     }
     tokv++;
     tokc--;
   }
   /***
    * option string should now be at the optional name, or if no name
    * is supplied, at the first layer size
    */
   if (!name && *tokv && !IIsInteger(tokv[0])) {
     name = tokv[0];
     tokv++;
     tokc--;
   }
   for (i=0;i<tokc;i++)
     if (!IIsInteger(tokv[i])) {
       IError("Expected <sizeof layer>, but found \"%s\"",tokv[i]);
       return;
     }
   if (input_shift && encode_input && input_shift!=encode_input)
     IErrorAbort("input-shift != encode-input");
   /***
    * make a net with tokc+1 layers - an extra layer for the bias
    */
  n_layers = tokc+1;
  if (verbosity>0) fprintf(dout,"Building network with %d layers\n",n_layers);
  if ((n=BuildNet(n_layers,name,1,NULL,builder))<0) return;
  net = neta[n];
  np = net->p;
  if (verbosity>0) fprintf(dout,"Building bias layer %d\n",0);
  BuildLayer(net,0,1,UNIT,NULL);
  for (i=1; i<n_layers; i++) {
    if (verbosity>0) fprintf(dout,"Building layer %d with %d units\n",i,atoi(tokv[i]));
    BuildLayer(net,i,atoi(tokv[i-1]),UNIT,NULL);
  }
  np->bias_layer_n   = 0;
  np->input_layer_n  = 1;	    /* these are no longer used in traversal */
  np->output_layer_n = n_layers-1;  /* and so can be set independently       */
  np->output_shift   = output_shift;
  np->input_shift    = input_shift;
  net->layer[np->bias_layer_n].auto_built = 1;
  if (squashing_function!=-1)
    np->unit_function = np->input_function = squashing_function;
  if (output_shift) {
    if ((np->output_width % output_shift) != 0)
      warn("build","shift does not partition output units");
    np->output_width = net->layer[np->output_layer_n].n_units / output_shift;
  } else
    np->output_width = net->layer[np->output_layer_n].n_units;
  if (input_shift) {
    if ((np->input_width % input_shift) != 0)
      warn("build","shift does not partition input units");
    np->input_width = net->layer[np->input_layer_n].n_units / input_shift;
  } else
    np->input_width = net->layer[np->input_layer_n].n_units;
  np->have_targets = 1;
  /*
   * Connect bias layer
   */
  if (autobias) {
    for (i=0; i<net->n_layers; i++) {
      if (i==np->input_layer_n && !input_bias) continue;
      if (verbosity>0) fprintf(dout,"Connecting bias layer to layer %d\n",i);
      ConnectComplete(n,np->bias_layer_n,n,i,LINK);
    }
  }
  if (encode_input) {
    int encode_size
      = (np->input_width * (np->input_shift?np->input_shift:1))/encode_input;
    if (!share_encoding) {
      for (i=np->input_layer_n+1; i<=encode_input; i++)
	Connect_(n,np->input_layer_n,n,i,TAG(LINK),
		 CONNECT_BLOCK|UNI_DIRECTIONAL,
		 (i-1)*encode_size,encode_size,0);
    } else {
      Connect_(n,np->input_layer_n,n,1,TAG(LINK),
	       CONNECT_BLOCK|UNI_DIRECTIONAL,
	       0,encode_size,0);
      /* the parasite call should get links to both input & bias layers */
      for (i=np->input_layer_n+2; i<=encode_input; i++)
	ConnectParasiteLayer(n,DFLT,(i-1)*encode_size,
			     n,i,
			     n,np->input_layer_n+1,DFLT,DFLT,DFLT,0);
    }
  }
  /*
   * set the bias unit to have value 1
   */
  SetBiases(net);
  /*
   * Randomize weights
   */
  if (restrict) ApplyToLinksOfNet(UNIT,LINK,UNIT,link->restrict=restrict;);
  if (seed) np->rand_seed = seed;
  if (verbosity>0)
    fprintf(dout,"Randomizing links between %g and %g with seed %d\n",
	   -np->rand_link_field,np->rand_link_field,np->rand_seed);
  if (np->rand_link_field==0.0)
    warn("build","links are all zero due to np.rand-link-field being zero");
  srandom(np->rand_seed);
  RandomizeAllLinks(net);
}
