#include <itf/option.h>
#include <itf/itf.h>
#include <sim/sim.h>
#include <lib/example.h>
#include "example-bp.h"
#include "prop-bp.h"

#define DB(x) x

#define EXTERNAL_INPUT(unit)	(unit)->dEdx

#define MAX_FP_TEMP	1.0	/* actual temp is final-temp + this */
#define AllowFP		(net->p->temp < MAX_FP_TEMP)

extern real GaussNoise();		/* example-bp.c */
extern real UnitNoise();		/* example-bp.c */
extern real output_cos_angle();		/* example-bp.c */
extern real euclidean_distance();	/* perform-bp.c */
extern real FreeEnergy();		/* energy-dbm.c */
extern char **calloc_2d();		/* perform-bp.c */

static int phase;
static real initial_temp;
real **fp = NULL;			/* fixedpoint storage */

static int display_spread;

#define SAVE_FP \
  (net->p->save_fp_freq && (net->p->epoch_n%net->p->save_fp_freq)==0)
#define RESTORE_FP \
  (net->p->save_fp_freq && fp!=NULL)

#define Clampable(layer)	\
  ((layer)->p->is_input_layer || (layer)->p->is_output_layer)

#define UpdatingClampedStates \
  (phase==NEGATIVE && (net->p->input_strength<1.0 || !IsInputIteration(net->p->max_t)))

#define ForEachClampedLayer	\
  ForEachLayer			\
    if (layer->p->clamped)


/******************************************************************************
 * Utilities
 */


void
  PrintUnitStates(net,temp)
NETWORK *net;
real temp;
{
  fprintf(dout,"%4.1f:",temp);
  ForEachUnitCompiled(UNIT) {
    fprintf(dout," %5.2f",UNIT_OUTPUT(unit));
  } EndForEach;
  fprintf(dout,"\n");
}


/******************************************************************************
 * Fixedpoint saving/restoring
 */

void
  SaveUnitStates(net,arr)
NETWORK *net;
real *arr;
{
  ForEachUnitCompiled(UNIT) {
    arr[unit_n] = UNIT_OUTPUT(unit);
  } EndForEach;
}

void
  RestoreUnitStates(net,arr)
NETWORK *net;
real *arr;
{
  ForEachUnitCompiled(UNIT) {
    UNIT_OUTPUT(unit) = arr[unit_n];
  } EndForEach;
}

int
  EnsureFixedpointStorage(net)
NETWORK *net;
{
  if (fp==NULL) {		/* saving positive phase */
    fp = (real**)calloc_2d(current_el->n_examples,net->p->n_units,sizeof(real));
    if (fp==NULL) {
      IOutOfMemory("EnsureFixedpoint","allocating fp (requested %d * %d)",
		   current_el->n_examples,net->p->n_units);
      return 0;
    }
  }
  return 1;
}

void
  SaveFixedpoint(net)
NETWORK *net;
{
  if (!EnsureFixedpointStorage(net)) return;
  SaveUnitStates(net,fp[net->p->example_n]);
}

void
  RestoreFixedpoint(net)
NETWORK *net;
{
  if (fp==NULL) { IError("RestoreFixedpoint called with fp = NULL");  return; }
  ForEachUnitCompiled(UNIT) {
    UNIT_OUTPUT(unit) = ShrinkOutput(fp[net->p->example_n][unit_n],
				     net->p->restore_strength);
  } EndForEach;
  ForEachClampedLayer {
    ForEachUnit(UNIT) {
      UNIT_OUTPUT(unit) = 
	UNIT_TARGET(unit) + UnitNoise(UNIT_TARGET(unit),net->p->input_noise);
    } EndForEach;
  } EndForEach;
}


/******************************************************************************
 * Fixedpoint settling
 */

void
  SetInitialStates(net)
NETWORK *net;
{
  ForEachLayer {
    if (IsBiasLayer) continue;
    if (layer->p->clamped) {
      ForEachUnit(UNIT) {
	UNIT_OUTPUT(unit) = 
	  UNIT_TARGET(unit) + UnitNoise(UNIT_TARGET(unit),net->p->input_noise);
      } EndForEach;
    } else {
      ForEachUnit(UNIT) {
	UNIT_OUTPUT(unit) = ScaleOutput(layer->p->initial_y);
      } EndForEach;
    }
  } EndForEach;
}

void
  MaskInput(net)
NETWORK *net;
{
  real s = net->p->input_strength;
  real p = net->p->mask_prob;
  real y;
  ForEachLayer {
    if (IsBiasLayer) continue;
    if (layer->p->clamped) {
      ForEachUnit(UNIT) {
	y = ScaleOutput(with_prob(p)) ;
	if (s < 1.0) {
	  y = (net->p->unit_function ? s*y : s*(y-0.5) + 0.5);
	  EXTERNAL_INPUT(unit) = ISQUASH(y);
	} else {
	  UNIT_OUTPUT(unit) = ScaleOutput(with_prob(p)) ;
	}
      } EndForEach;
    }
  } EndForEach;
} 

void
  SetExternalInput(net,restore)
NETWORK *net;
int restore;
{
  real y,s;
  ForEachLayer {
    if (   ( layer->p->clamped && UpdatingClampedStates && (s=net->p->input_strength))
	|| (!layer->p->clamped && restore               && (s=net->p->restore_strength)))
      ForEachUnit(UNIT) {
	y = (net->p->unit_function ? s*UNIT_OUTPUT(unit) : s*(UNIT_OUTPUT(unit)-0.5) + 0.5);
	EXTERNAL_INPUT(unit) = ISQUASH(y);
	UNIT_OUTPUT(unit) = ScaleOutput(layer->p->initial_y);
      } EndForEach;
  } EndForEach;
}

void
  AddExternalInput(net)
NETWORK *net;
{
  NET_PARAMS *np = net->p;
  real noise;
  if (!IsInputIteration(np->max_t)) return;
  if (np->unit_noise!=0.0) {
    noise = np->unit_noise * (AllowFP ? np->temp/MAX_FP_TEMP : 1.0);
    ForEachUnitCompiled(UNIT) {
      unit->x += EXTERNAL_INPUT(unit) + GaussNoise(noise);
    } EndForEach;
  } else {
    ForEachUnitCompiled(UNIT) {
      unit->x += EXTERNAL_INPUT(unit);
    } EndForEach;
  }
}

void
  ClearInput(net)
NETWORK *net;
{
  ForEachLayer {
    ForEachUnit(UNIT) {
      unit->x = EXTERNAL_INPUT(unit) = 0.0;
    } EndForEach;
  } EndForEach;
}

void
  InitializeStates(net,restore)
NETWORK *net;
int restore;
{
  real y;
  initial_temp = net->p->initial_temp;
  net->p->temp = initial_temp - net->p->final_temp;
  net->p->max_t = 1;
  if (restore) RestoreFixedpoint(net);
  else         SetInitialStates(net);
  SetExternalInput(net,restore);
}

void
  SumInput(net)
NETWORK *net;
{
  if (net->p->hebb_scale==0.0) {
    ForEachLinkCompiled(LINK,UNIT) {
      d_unit->x += UNIT_OUTPUT(s_unit) * LINK_WEIGHT(link);
      s_unit->x += UNIT_OUTPUT(d_unit) * LINK_WEIGHT(link);
    } EndForEach;
  } else {
    ForEachLinkCompiled(LINK,UNIT) {
      d_unit->x +=         UNIT_OUTPUT(s_unit)     *LINK_WEIGHT(link) + 
	net->p->hebb_scale*UNIT_HEBB_OUTPUT(s_unit)*LINK_HEBB_WEIGHT(link);
      s_unit->x +=         UNIT_OUTPUT(d_unit)     *LINK_WEIGHT(link) +
	net->p->hebb_scale*UNIT_HEBB_OUTPUT(d_unit)*LINK_HEBB_WEIGHT(link);
    } EndForEach;
  }
}

int UpdateStates(net,show)
NETWORK *net;
int show;
{
  int  at_fp = AllowFP;
  real tol = net->p->fp_tolerance;
  real d  = net->p->unit_damping,  di = 1.0-d;
  real temp  = net->p->temp + net->p->final_temp;
  real max_change = 0.0;
  real change,old_y;
  char command_str[256];
  ForEachLayer {
    if (IsBiasLayer || (layer->p->clamped && !UpdatingClampedStates)) continue;
    if (phase==POSITIVE && !net->p->settle_positive) {
      ForEachUnit(UNIT) {
	UNIT_OUTPUT(unit) = SQUASH(unit->x);
      } EndForEach;
      at_fp = TRUE;
    } else {
      ForEachUnit(UNIT) {
	old_y             = UNIT_OUTPUT(unit);
	UNIT_OUTPUT(unit) = d*UNIT_OUTPUT(unit) + di*SQUASH(unit->x/temp);
	unit->x           = 0.0;
	change		  = fabs(old_y-UNIT_OUTPUT(unit));
	if (change > max_change) max_change = change;
	if (change > tol) at_fp = FALSE;
      } EndForEach; /* Unit */
    } /* settle */
  } EndForEach; /* Layer */
  if (show&SHOW_PRINT) PrintUnitStates(net,temp);
  if (show&SHOW_DISPLAY) {
    if (net->p->max_t==1) fprintf(dout,"clear\n");
    sprintf(command_str,
	    "display -weights +states -clear +repetitive -box %sspread +label \"Iteration %3d: T = %6.3f   E = %10.3f   dY = %5.3f\"",
	    (display_spread?"+":"-"),net->p->max_t,temp,FreeEnergy(net,FALSE,FALSE),max_change);
    IDoCommandLine(command_str);
  }
  if (net->p->max_t==np->max_iterations) return TRUE;
  net->p->max_t++;
  if (net->p->max_t==np->mask_onset) MaskInput(net);
  net->p->temp *= net->p->temp_decay;
  return(at_fp);
}
 
void
  UpdateGradients(net)
NETWORK *net;
{
  int nclamped = 0;
  real faze = -1.0;
  if (phase==POSITIVE) {
    ForEachLayer {
      if (Clampable(layer)) nclamped++;
    } EndForEach;
    faze = ((real) nclamped);
  }
  ForEachLinkCompiled(LINK,UNIT) {
    link->dEdw -= faze * UNIT_OUTPUT(s_unit) * UNIT_OUTPUT(d_unit);
  } EndForEach;
}

void UpdateHebbWeights(net)
NETWORK *net;
{
  NET_PARAMS *np = net->p;
  ForEachLinkCompiled(LINK,UNIT) {
    LINK_HEBB_WEIGHT(link) =
      np->hebb_damping*LINK_HEBB_WEIGHT(link) +
      (1.0-np->hebb_damping)*UNIT_HEBB_OUTPUT(s_unit)*UNIT_HEBB_OUTPUT(d_unit);
  } EndForEach;
}

void
  SettleToFixedpoint(net,show,update,save,restore)
NETWORK *net;
int show,update,save,restore;
{
  NET_PARAMS *np = net->p;
  if (!net->p->compiled) CompileNet(net);
  ClearInput(net);
  InitializeStates(net,restore);
  if (show&SHOW_PRINT) PrintUnitStates(net,initial_temp);
  do {
    AddExternalInput(net);
    SumInput(net);
  } while (!UpdateStates(net,show));
  if (save) SaveFixedpoint(net);
  if (update) UpdateGradients(net);
}


/******************************************************************************
 * Compute Errors
 */

void
  ComputeError(net)
NETWORK *net;
{
  /*
   * done within BackwardProp() by AccumulateError();
   */
}

void
  AccumulateError(net)
NETWORK *net;
{
  NET_PARAMS *np = net->p;
  if (np->output_criteria) {
    if (OutputMeetsCriteria(net,0)!=1) np->ex_n_errors++;
  } else {
    if (net->layer[np->input_layer_n].p->clamped) {
      np->sum_cos_angle += output_cos_angle(net);
      if (np->therapy) PrintExamplePerformance(net);
    }
    ForEachLayer {
      if (Clampable(layer) && !layer->p->clamped) {
	ForEachUnit(UNIT) {
	  np->ex_sumsq_errors += UNIT_ERROR(unit,NULL);
	  if (IsBadUnit(unit,NULL)) np->ex_n_errors++;
	} EndForEach; /* Unit */
      } /* clamped */
    } EndForEach; /* Layer */
  }
  np->total_n_errors += np->ex_n_errors;
  np->sumsq_errors   += np->ex_sumsq_errors;
}


/*****************************************************************************
 * ForwardProp() and BackwardProp()
 */

void
  ForwardProp(net)		/* positive phase */
NETWORK *net;
{
  phase = POSITIVE;
  net->layer[net->p->input_layer_n].p->is_input_layer = TRUE;
  net->layer[net->p->output_layer_n].p->is_output_layer = TRUE;
  ForEachLayer {
    if (Clampable(layer)) layer->p->clamped = TRUE;
  } EndForEach;
  SettleToFixedpoint(net,FALSE,TRUE,SAVE_FP,FALSE);
  ForEachLayer {
    layer->p->clamped = FALSE;
  } EndForEach;
}

void
  BackwardProp(net)		/* negative phase */
NETWORK *net;
{
  NET_PARAMS *np = net->p;
  phase = NEGATIVE;
  np->ex_n_errors = 0;
  np->ex_sumsq_errors = 0;
  if (RESTORE_FP && !net->p->example_n) fprintf(dout,"+");
  ForEachLayer {
    if (layer->p->is_input_layer) {
      layer->p->clamped = TRUE;
      SettleToFixedpoint(net,FALSE,TRUE,FALSE,RESTORE_FP);
      AccumulateError(net);
      layer->p->clamped = FALSE;
    }
  } EndForEach;
  if (np->hebb_scale>0.0) UpdateHebbWeights(net);
}

int PerformNet_DBM(net)
NETWORK *net;
{
  NET_PARAMS *np = net->p;
  phase = NEGATIVE;
  np->ex_n_errors = 0;
  np->ex_sumsq_errors = 0;
  net->layer[np->input_layer_n].p->clamped = TRUE;
  SettleToFixedpoint(net,FALSE,FALSE,FALSE,FALSE);
  net->layer[np->input_layer_n].p->clamped = FALSE;
  return 0;
}


/******************************************************************************
 * Commands
 */

int
  command_forward(tokc,tokv)
int tokc;
char *tokv[];
{
  int save = FALSE, show = FALSE, print = FALSE, display = FALSE;
  int update = FALSE, example_n = -1;
  char *example_str = NULL;
  static struct OPTION options[] =
    {
      {OPTPRES,"save",	FALSE},
      {OPTPRES,"ex",	FALSE},
      {OPTSIGN,"print",	FALSE},
      {OPTSIGN,"display",FALSE},
      {OPTSIGN,"spread",FALSE},
      {OPTSIGN,"update",FALSE},
      {OPTLAST,NULL,	FALSE}
    };
  IUsage("[%pex <example>] [%psave] [%sprint] [%sdisplay]\
 [%sspread] [%supdate]");
  if (GiveHelp(tokc)) {
    ISynopsis("Perform forward (positive) pass.");
    IHelp(IHelpArgs,
"Run the forward pass, possibly on a particular example.",
"ARGUMENTS",
"%psave         : save fixedpoint after running.",
"%pex <example> : example to run (<int> or <comment>)",
"%sprint      : print units states during settling (default %n)",
"%sdisplay    : display units states during settling (default %n)",
"%sspread     : do (not) spread layers in state display (default %y)",
"%supdate     : update weight gradients (default %n)",
	  NULL);
    return;
  }
  if (!NetExists(net)) { IError("current net is null");  return; }
  net->layer[net->p->input_layer_n].p->is_input_layer = TRUE;
  net->layer[net->p->output_layer_n].p->is_output_layer = TRUE;
  display_spread = TRUE;
  /***
   * process options
   */
  tokv++;	/* ignore command name */
  while (IsOption(tokv[0])) {
    if (OptionMatch(tokv[0],"ex",options,NULL)) {
      if (IIsInteger(tokv[1])) example_n = atoi(*++tokv);
      else 		       example_str = *++tokv; 
      tokc--;
      if (!LoadExample(net,example_n,example_str)) return;
    } else if (OptionMatch(tokv[0],"save",options,&save)) {
    } else if (OptionMatch(tokv[0],"spread",options,&display_spread)) {
    } else if (OptionMatch(tokv[0],"print",options,&print)) {
      show |= print;
    } else if (OptionMatch(tokv[0],"display",options,&display)) {
      show |= display<<1;
    } else if (OptionMatch(tokv[0],"update",options,&update)) {
    } else {
      IError("option not understood: \"%s\"",tokv[0]);
      return;
    }
    tokv++;
    tokc--;
  }
  phase = POSITIVE;
  ForEachLayer {
    if (Clampable(layer)) layer->p->clamped = TRUE;
  } EndForEach;
  SettleToFixedpoint(net,show,update,save,FALSE); 
  ForEachLayer {
    layer->p->clamped = FALSE;
  } EndForEach;
}


int
  command_backward(tokc,tokv)
int tokc;
char *tokv[];
{
  int restore = FALSE, show = FALSE, print = FALSE, display = FALSE;
  int update = FALSE, all = TRUE, hebb = FALSE, example_n = -1;
  char *example_str = NULL;
  static struct OPTION options[] =
    {
      {OPTPRES,"restore",FALSE},
      {OPTPRES,"ex",	FALSE},
      {OPTSIGN,"print",	FALSE},
      {OPTSIGN,"display",FALSE},
      {OPTSIGN,"spread", FALSE},
      {OPTSIGN,"update",FALSE},
      {OPTSIGN,"all",	FALSE},
      {OPTSIGN,"hebb",	FALSE},
      {OPTLAST,NULL,FALSE}
    };
  IUsage("[%pex <example>] [%prestore] [%sprint] [%sdisplay]\
 [%sspread] [%supdate] [%sall] [%shebb]");
  if (GiveHelp(tokc)) {
    ISynopsis("Perform backward (negative) pass.");
    IHelp(IHelpArgs,
"Run the backward pass, possibly on a particular example.",
"ARGUMENTS",
"%prestore      : restore (positive) fixedpoint before running.",
"%pex <example> : example to run (<int> or <comment>)",
"%sprint      : print units states during settling (default %n)",
"%sdisplay    : display units states during settling (default %n)",
"%sspread     : do (not) spread layers in state display (default %n)",
"%supdate     : update weight gradients (default %n)",
"%sall        : run all input layers (default %y)",
"%shebb       : update hebb weights (default %n)",
	  NULL);
    return;
  }
  if (!NetExists(net)) { IError("current net is null"); return; }
  net->layer[net->p->input_layer_n].p->is_input_layer = TRUE;
  net->layer[net->p->output_layer_n].p->is_output_layer = TRUE;
  display_spread = FALSE;
  /***
   * process options
   */
  tokv++;	/* ignore command name */
  while (IsOption(tokv[0])) {
    if (OptionMatch(tokv[0],"ex",options,NULL)) {
      if (IIsInteger(tokv[1])) example_n = atoi(*++tokv);
      else 		       example_str = *++tokv; 
      tokc--;
      if (!LoadExample(net,example_n,example_str)) return;
    } else if (OptionMatch(tokv[0],"restore",options,&restore)) {
    } else if (OptionMatch(tokv[0],"spread",options,&display_spread)) {
    } else if (OptionMatch(tokv[0],"all",options,&all)) {
    } else if (OptionMatch(tokv[0],"hebb",options,&hebb)) {
    } else if (OptionMatch(tokv[0],"print",options,&print)) {
      show |= print;
    } else if (OptionMatch(tokv[0],"display",options,&display)) {
      show |= display<<1;
    } else if (OptionMatch(tokv[0],"update",options,&update)) {
    } else {
      IError("option not understood: \"%s\"",tokv[0]);
      return;
    }
    tokv++;
    tokc--;
  }
  phase = NEGATIVE;
  ForEachLayer {
    layer->p->clamped = FALSE;
  } EndForEach;
  ForEachLayer {
    if (layer->p->is_input_layer && (all || IsInputLayer)) {
      layer->p->clamped = TRUE;
      SettleToFixedpoint(net,show,update,FALSE,restore);
      layer->p->clamped = FALSE;
    }
  } EndForEach;
  if (hebb) UpdateHebbWeights(net);
}


int
  command_set_dbm_gain(tokc,tokv)
int tokc;
char *tokv[];
{
  IUsage("");
  if (GiveHelp(tokc)) {
    ISynopsis("set gain on links of DBM.");
    IHelp(IHelpArgs,
"Set gain on links of a DBM",
	  NULL);
    return;
  }
  if (!NetExists(net)) {
    IError("current net is null");
    return;
  }
  fprintf(dout,"Setting gain to 1/fan-in (DBM)\n");
  ForEachUnitCompiled(UNIT) {
    unit->x = 0.0;
  } EndForEach;
  ForEachLinkCompiled(LINK,UNIT) {
    s_unit->x++;
    d_unit->x++;
  } EndForEach;
  ForEachLinkCompiled(LINK,UNIT) {
    link->gain = 2.0/(s_unit->x + d_unit->x);
  } EndForEach;
}

