/***
 * This update record is intended to allow the checking of gradients.
 * 
 * The command "check-gradients" does the following:
 * -- first calculate the error & derivatives
 * for each example {
 *   ForwardProp(net);
 *   BackwardProp(net);
 * }
 * store E in update.original_E
 * for each link & time step {
 *   update.link_num is set to the link number
 *   update.t is set to the time
 *   UpdateOneWeight(net) - Which just calculate an update for a link, and
 *                        - store it in new_w (i.e. don't change any links)
 *                        - The following fields are used:
 *			  - grad       - if 1 do a gradient movement, otherwise
 *                                     - proportional to w
 *                        - delta      - the actual change made
 *                        - dEdw       - the local gradient
 *                        - old_w      - store the old weight
 *                        - new_w      - the new weight
 *                        - link       - pointer to the link to be changed
 *
 *   ForwardProp(net)     - Which should load the new weight at the
 *                          appropriate time, using the pointer provided
 *   the resulting E (in np->sumsq_errors) is stored in update.E_prime,
 *   and the ratio of expected change to actual change is calculated
 *   and stored
 * }
 */
typedef struct UPDATE {	/* BIND */
  int	grad;		/* default: 0 if 1 do gradient descent */
  int	link_n;
  real	delta;		/* actual change made to weight */
  real	new_w;		/* the new weight to use */
  real	old_w;		/* the old weight */
  real	dEdw;
  real	E;
  real	E_prime;
  real	dE;
  real	ratio;
  LINK	*link;
  int	warned;
#ifdef TIME_UPDATE_CHECKING
  int	t;
#endif
};

extern struct TRACE check_update; /* BIND */
extern struct UPDATE update;	/* BIND */

#define update_momentum(link,alpha)					\
  (link)->momentum = (alpha)*(link)->momentum + (link)->dEdw

#define decay_weight(link,decay)					\
  (link)->dEdw += LINK_WEIGHT(link)*(decay)

#define update_weight(link,update)					\
  if        ((link)->restrict == 0) {					\
    LINK_WEIGHT(link) -= (update);					\
  } else if ((link)->restrict == -1) {	 				\
    if ((LINK_WEIGHT(link)-=(update))>0.0) LINK_WEIGHT(link) = 0.0;	\
  } else								\
    if ((LINK_WEIGHT(link)-=(update))<0.0) LINK_WEIGHT(link) = 0.0;

#define update_dbd_gain(link,np) 					\
  if ((np)->delta_bar_delta) {						\
    if (((link)->dEdw*((link)->delta_bar=(link)->dEdw+(np)->beta*(link)->delta_bar))>0) \
      (link)->gain = min((link)->gain+(np)->gain_incr,(np)->max_gain); 	\
    else 								\
      (link)->gain *= (np)->gain_decr; 					\
  }

#define update_stats(link,np,update) 					\
{ 									\
  (np)->sum_grad_by_update  += (link)->dEdw * update; 			\
  (np)->sum_old_grad_by_new += (link)->previous_dEdw * (link)->dEdw; 	\
  (np)->sum_gradient        += fabs((link)->dEdw); 			\
  (np)->sum_update          += fabs(update); 				\
  (np)->sumsq_gradient      += sqr((link)->dEdw); 			\
  (np)->sumsq_update        += sqr(update); 				\
  (np)->sum_weight          += fabs(LINK_WEIGHT(link)); 		\
  (np)->sumsq_weight        += sqr(LINK_WEIGHT(link)); 			\
}

#define clear_dEdw(link) 						\
{ 									\
  (link)->previous_dEdw      = (link)->dEdw; 				\
  (link)->dEdw               = 0; 					\
}
