/*
 * $Id: save.c,v 1.3 91/03/06 21:21:44 xerion Exp $
 * 
 * $Log:	save.c,v $
 * Revision 1.3  91/03/06  21:21:44  xerion
 * added MarkToSave back into the file. Don't know how
 * or why it was ever taken out.
 * 
 * Revision 1.2  90/12/14  15:01:41  drew
 * Documented and cleaned up.
 * 
 * Revision 1.1  90/12/07  12:00:58  drew
 * Initial revision
 * 
 */

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

/**********************************************************************
 *	File name:   save.c
 *	Description: contains the functions used to save and load networks
 *
 *	Exported functions:
 *		int	command_save(tokc,tokv)
 *		int	command_use_net(tokc,tokv)
 *		int	command_link_list(tokc,tokv)
 *		int	command_network(tokc,tokv)
 *
 *	Exported variables:
 *		none
 *
 **********************************************************************/

#include <itf/option.h>
#include <itf/itf.h>
#include "sim-dummy.h"
#include <sim.h>

/**********************************************************************
 *	Name:         SaveNetworkHeader
 *	Description:  writes the header description of a network to `dout'
 *			The header consists of a command to build the
 *			skeleton structure for the network being saved
 *			when it is read in again.
 *	Parameters:   NETWORK*	tnet - the net being saved
 *	Return Value: none
 **********************************************************************/
static int	SaveNetworkHeader(tnet)
NETWORK*	tnet;
{
  struct NETWORK	*old_net = net;
  struct NET_PARAMS	*old_np  = np;

  net = tnet;
  np  = net->p;

  if (!NetExists(net)) {
    IError("net is null");
    return;
  }
  fprintf(dout,"network ");
  IPutQuotedString(net->builder, '"', dout);
  fprintf(dout," %s %d %d",
	  net->name, net->n_layers, net->n_links);
  ForEachLayer {
    fprintf(dout," %d",layer->n_units);
  } EndForEach;
  fprintf(dout,"\n");
  IShowObject("net",SHOW_SAVE);
  net = old_net;
  np = old_np;
}
/**********************************************************************/

/**********************************************************************
 *	Name:         SaveLinkList
 *	Description:  writes the linked list of weights for a network 
 *			to dout
 *	Parameters:   NETWORK	*tnet - a pointer to the network to save
 *	Return Value: int	SaveLinkList - the value of SaveWeights
 **********************************************************************/
static int	SaveLinkList(tnet)
NETWORK 	*tnet;
{
  extern int	SaveWeights() ;

  fprintf(dout,"link-list %d\n",tnet->n_links);
  return SaveWeights(tnet);
}
/**********************************************************************/

/***********************************************************************
 *	Name:		MarkToSave
 *	Description:	marks a structure field to be saved
 *	Parameters:	
 *		char	*struct_field - a pointer to the field
 *	Return Value:	
 *		none
 ***********************************************************************/
int	MarkToSave(struct_field)
  char *struct_field;
{
  ISetFeatures(struct_field,SAVE_FEATURE,1);
}
/**********************************************************************/


/**********************************************************************
 *	Name:         SaveNetwork
 *	Description:  saves the network passed in to it
 *	Parameters:   NETWORK	*tnet - a pointer to the network to save
 *	Return Value: int	SaveNetwork - +1 on success, 0 on failure
 **********************************************************************/
static int	SaveNetwork(tnet)
NETWORK		*tnet ;
{
  if (tnet!=NULL && NetExists(tnet)) {
    SaveNetworkHeader(tnet);
    if (SaveLinkList(tnet)) {
      IError("Failed to save weights for net \"%s\"",tnet->name);
      return 0;
    }
  }
  return 1 ;
}
/**********************************************************************/

/**********************************************************************
 *	Name:         command_save
 *	Description:  writes the header and weights of a network to dout
 *	Parameters:   int tokc     - the number of command line tokens
 *		      char *tokv[] - the vector of tokens
 *	Return Value: none
 **********************************************************************/
int		command_save(tokc,tokv)
  int		tokc;
  char		*tokv[];
{
  int		i, allnets = 0, parameters = -1 ;
  char		*name = NULL;
  NETWORK	*thisnet;
  struct ARG_DESC *argd;

  argd = StartArgs(tokv[0]);
  Args(argd, "[<network> %s %?]", &name,
       "save the network with this name, rather than the current network.");
  Args(argd, "[-Pallnets %P %?]", &allnets,
       "save all of the networks (default is save only current)") ;
  Args(argd, "[-Pparameters %P %?]", &parameters, 
       "save the parameters block (default is nosave unless all nets are being saved)") ;
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("print out the essentials of a network");
    IHelp(tokc, tokv[0], NULL, synopsis,
       "Print out the weights in a network.",
       "To get this output into a file, use redirection, e.g.:",
       "  save > filename\n",
       "CUSTOMIZATION",
       "This command calls the function `SaveWeights' to write out the weights.",
       "This command can be customized by writing your own version of",
       "SaveWeights()",
	  NULL);
    return;
  }
  (void) ParseArgs(argd, tokc, tokv, 0);

  if (name && allnets)
    IErrorAbort("Cannot give both name and allnets option");

  if (allnets) {
    for (i=0;i<n_nets;i++)
      if (SaveNetwork(neta[i]) == 0) return ;
  } else {
    if (name) {
      thisnet = (NETWORK*)IGetObjectPtr("NETWORK", name);
      if (thisnet==NULL) {
	IError("network \"%s\" not found", name);
	return ;
      }
    } else
      thisnet = net;
    
    if (!NetExists(thisnet)) {
      IError("network \"%s\" is null",(thisnet!=NULL?thisnet->name:"no net"));
      return;
    }
    if (SaveNetwork(thisnet) == 0) return ;
  }

  if (parameters < 0)
    parameters = allnets ;

  if (parameters)
    IShowObject("p",SHOW_SAVE);
}
/**********************************************************************/

/**********************************************************************
 *	Name:         command_network
 *	Description:  this command is called when reading in a save
 *			file. It should be called from the first line
 *			of the save file (or before every net in the
 *			save file). It is not designed as an interactive
 *			command. The SaveNetworkHeader function generates
 *			the line which calls this command.
 *	Parameters:   int tokc     - the number of command line tokens
 *		      char *tokv[] - the vector of tokens
 *	Return Value: none
 **********************************************************************/
int	command_network(tokc,tokv)
  int	tokc;
  char	*tokv[];
{
  int	net_n,i,n;
  char	*name; /* the net name */

  IUsage("<build-command> <name> <#layers> <#links> <layer0 size> ...");
  if (GiveHelp(tokc)) {
    ISynopsis("ensures that a net of the given size exists");
    IHelp(IHelpArgs,
	  "This command is used for reading a network from a file. It",
	  "ensures that the necessary network structure exists for storing",
	  "the saved network from the file. A `network' command is generated",
	  "by the `save' command and written to the save file.  It is",
	  "executed when the save file is read in.\n",
	  "This command is not intended to be used interactively.",
	  NULL);
    return;
  }
  if (tokc<5) {
    IError(IPrintUsage(tokv[0],usage));
    return;
  }
  if (!IIsLegalVarName(tokv[2])) {
    IExpectedToken("<network-name>",tokv[2],NULL);
    return;
  }
  name = tokv[2];
  /* Find the net.  If it does not exist, then build it using
   * the build command from the command line
   */
  net_n = FindOrBuildNetSkel(name);
  if (net_n>=0)	net = neta[net_n];
  else		net = NULL;
  if (!NetExists(net)) {
    fprintf(dout,"Building new net \"%s\" ...\n",name);
    if (IDoCommandLine(tokv[1])!=0) return; /* there was an error! */
  } else
    fprintf(dout,"Found net \"%s\", checking similarity...\n",name);

  np = net->p;
  if (strcmp(net->builder,tokv[1]))
    fprintf(dout,
       "Warning: builder strings are different:\n%-13s %s\n%-13s %s\n%s\n",
       "existing net:",net->builder,"saved net:",tokv[1],
       "Continuing anyway...");
    
  /***
   * Check the # of layers, the # of links, and the size of each layer
   */
  n = atoi(tokv[3]);
  if (net->n_layers!=n) {
    IError("Mismatch: net has %d layers, but %s were requested",
	   net->n_layers,tokv[3]);
    return;
  }
  if (net->n_links!=atoi(tokv[4])) {
    IError("Mismatch: net has %d links, but %s were requested",
	   net->n_links,tokv[4]);
    return;
  }
  if (tokc!=5+n) {
    IError("number of arguments does not match number of layers");
    return;
  }
  for (i=0;i<net->n_layers;i++) {
    if (atoi(tokv[i+5])!=net->layer[i].n_units) {
      IError("wrong number of units in layer %d (net has %d, command has %s)",
	     i,net->layer[i].n_units,tokv[i+5]);
      return;
    }
  }
}
/**********************************************************************/

/**********************************************************************
 *	Name:         command_link_list 
 *	Description:  This function takes as argument the number of links
 *			to fill in and then reads in the values to
 *			store in the list. It is not an interactive
 *			command. It is called from save files.
 *	Parameters:   int tokc     - the number of command line tokens
 *		      char *tokv[] - the vector of tokens
 *	Return Value: none
 **********************************************************************/
int	command_link_list(tokc,tokv)
int	tokc ;
char 	*tokv[];
{
  extern int	  ReadWeights() ;
  int		  nLinks = -1 ;
  struct ARG_DESC *argd ;

  argd = StartArgs(tokv[0]);
  Args(argd, "(<num> %n %?)", &nLinks, 
       "the number of links to be read in.") ;
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("set the current net");
    IHelp(tokc, tokv[0], NULL, synopsis,
	  "This command is just intended to be used in save files,",
	  "it calls a function `ReadWeights' to read the weights of",
	  "the network from the input.",
	  NULL);
    return;
  }
  (void) ParseArgs(argd, tokc, tokv, 0);
  
  if (!NetExists(net)) {
    IError("current net is null");
    return;
  }
  if (net->n_links != nLinks) {
    IError("wrong number of links: net has %d, input has %d",
		  net->n_links,atoi(tokv[1]));
    return;
  }
  fprintf(dout,"Reading weights...\n");
  if (ReadWeights(net)>0) {
    IError("error occurred reading weights");
    return;
  }
  return;
}
/**********************************************************************/

/**********************************************************************
 *	Name:         command_use_net
 *	Description:  command for telling the simulator which net to
 *			use or for clearing the current net.
 *			It resets the variables net, np and lp.
 *	Parameters:   int tokc     - the number of command line tokens
 *		      char *tokv[] - the vector of tokens
 *	Return Value: none
 **********************************************************************/
int		command_use_net(tokc,tokv)
  int		tokc;
  char		*tokv[];
{
  int		net_n = -1 ;
  int		new = 0;
  char		*name = NULL;
  struct ARG_DESC *argd;

  argd = StartArgs(tokv[0]);
  Args(argd, "[<num> %n %?]", &net_n, 
       "use the network at  position <num> in the network array.");
  Args(argd, "[<name> %s %?]", &name, 
       "use the network with name <name>, if it does not exist, create it.");
  Args(argd, "[-Pnew %P %?]", &new, 
       "create or find an unused location in the network array, and use it.") ;
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("set the current net");
    IHelp(tokc, tokv[0], NULL, synopsis,
	  "Set the variables `net', `np' (net parameters), and `lp' (layer",
	  "parameters) to new values.  All commands that work on a network",
	  "use the variable `net' as the current network.\n",
	  "Only one option may be given to use-net at a time.  If no arguments",
	  "are given, the name and number of the current net are printed.",
	  NULL);
    return;
  }
  (void) ParseArgs(argd, tokc, tokv, 0);

  /* check command line for errors */
  if (net_n >= 0 && name)
    IErrorAbort("Cannot give both number and name");
  if (net_n >= 0 && new)
    IErrorAbort("Cannot give both number and \"new\"");
  if (name && new)
    IErrorAbort("Cannot give both name and \"new\"");

  if (net_n >= 0) {
    /* network # */
    if (net_n >= n_nets) (void) LengthenNetA();
    if (net_n<0 || net_n>=n_nets) {
      IError ("%d is not in the range of the current network array [%d-%d]",
	      net_n,0,n_nets-1);
      return;
    }
  } else if (name!=NULL) {
    /* network name */
    net_n = FindOrBuildNetSkel(tokv[1]);
    if (net_n < 0) return;
  } else if (new) {
    /* no arguments */
    net_n = FindNullNetA();
    if (net_n<0) {
      IError("can't create any more net slots in neta");
      return;
    }
  }

  /* set the current net */
  if (net_n >= 0) {
    net = neta[net_n];
    if (NetExists(net)) UseNet(net);
  }

  /* print out the net */
  if (net==NULL)
    fprintf(dout, "Current net is NULL\n");
  else {
    if (!NetExists(net))
      fprintf(dout,"Warning: current net is now null\n");
    fprintf(dout, 
	    "Current net is called \"%s\", and is at position %d in neta\n",
	    net->name, net->my_n);
  }
}
/**********************************************************************/


